"""Test we don't break TurboGears' usage of our Python API."""

from __future__ import annotations

import dataclasses
import wsgiref.util
from pathlib import Path
from typing import Any

import pytest

from kajiki import i18n

DATA = Path(__file__).resolve().parent.parent / "data"
GOLDEN = DATA / "golden"


@pytest.fixture
def tg2():
    """Import tg2 (if available), and cleanup its mess."""
    orig_gettext = i18n.gettext
    try:
        import tg
    except ImportError:
        pytest.skip("TurboGears not installed")
    else:
        yield tg
    finally:
        i18n.gettext = orig_gettext


@pytest.fixture
def tg2_app(tg2):
    """Create a new tg2 app with defaults for testing.

    Returns:
        A WSGI application.
    """

    class RootController(tg2.TGController):
        @tg2.expose("kajiki_test_data.kitchensink")
        def index(self):
            return {}

    config = tg2.MinimalApplicationConfigurator()
    config.update_blueprint(
        {
            "root_controller": RootController(),
            "renderers": ["kajiki"],
            "templating.kajiki.template_extension": ".html",
        }
    )
    return config.make_wsgi_app()


@dataclasses.dataclass
class Response:
    """Encapsulates a response for testing."""

    status: str
    headers: list[tuple[str, str]]
    body: str


def app_request(app: Any, path: str = "/", method: str = "GET") -> Response:
    response_status = ""
    response_headers = []

    def start_response(status, headers):
        nonlocal response_status
        nonlocal response_headers
        response_status = status
        response_headers = headers

    environ = {}
    wsgiref.util.setup_testing_defaults(environ)
    environ["REQUEST_METHOD"] = method
    environ["PATH_INFO"] = path

    response_body = b"".join(app(environ, start_response))

    return Response(
        status=response_status,
        headers=response_headers,
        body=response_body,
    )


def test_smoke_tg2(tg2_app):
    """Do a basic smoke test of using kajiki renderer on tg2."""
    response = app_request(tg2_app, "/")
    assert response.status == "200 OK"
    assert response.body == (GOLDEN / "kitchensink1.html").read_bytes()



import importlib
import pathlib
import sys

import pytest


@pytest.fixture(scope="session", autouse=True)
def add_test_data_module():
    here = pathlib.Path(__file__).parent
    data_package = here / "data" / "__init__.py"
    spec = importlib.util.spec_from_file_location("kajiki_test_data", str(data_package))
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)



from unittest import TestCase

import kajiki


class TestBasic(TestCase):
    def setUp(self):
        class Tpl:
            @kajiki.expose
            def __main__():
                yield "Hello,"
                yield name  # noqa: F821
                yield "\n"

        self.tpl = kajiki.Template(Tpl)

    def test_basic(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert rsp == "Hello,Rick\n", rsp


class TestSwitch(TestCase):
    def setUp(self):
        class Tpl:
            @kajiki.expose
            def __main__():
                for i in range(2):
                    yield local.__kj__.escape(i)  # noqa: F821
                    yield " is "
                    local.__kj__.push_switch(i % 2)  # noqa: F821
                    if local.__kj__.case(0):  # noqa: F821
                        yield "even\n"
                    else:
                        yield "odd\n"

        self.tpl = kajiki.Template(Tpl)

    def test_basic(self):
        rsp = self.tpl().render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestFunction(TestCase):
    def setUp(self):
        class Tpl:
            @kajiki.expose
            def evenness(self):
                if self % 2 == 0:
                    yield "even"
                else:
                    yield "odd"

            @kajiki.expose
            def __main__():
                for i in range(2):
                    yield local.__kj__.escape(i)  # noqa: F821
                    yield " is "
                    yield evenness(i)  # noqa: F821
                    yield "\n"

        self.tpl = kajiki.Template(Tpl)

    def test_basic(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestCall(TestCase):
    def setUp(self):
        class Tpl:
            @kajiki.expose
            def quote(self, speaker):
                for i in range(2):
                    yield "Quoth "
                    yield speaker
                    yield ', "'
                    yield self(i)
                    yield '."\n'

            @kajiki.expose
            def __main__():
                @__kj__.flattener.decorate  # noqa: F821
                def _kj_lambda(n):
                    yield "Nevermore "
                    yield local.__kj__.escape(n)  # noqa: F821

                yield quote(_kj_lambda, "the raven")  # noqa: F821
                del _kj_lambda

        self.tpl = kajiki.Template(Tpl)

    def test_basic(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert rsp == 'Quoth the raven, "Nevermore 0."\n' 'Quoth the raven, "Nevermore 1."\n', rsp


class TestImport(TestCase):
    def setUp(self):
        class LibUndec:
            @kajiki.expose
            def evenness(self):
                if self % 2 == 0:
                    yield "even"
                else:
                    yield "odd"

            @kajiki.expose
            def half_evenness(self):
                yield " half of "
                yield local.__kj__.escape(self)  # noqa: F821
                yield " is "
                yield evenness(self / 2)  # noqa: F821

        lib = kajiki.Template(LibUndec)

        class Tpl:
            @kajiki.expose
            def __main__():
                simple_function = lib(dict(globals()))
                for i in range(4):
                    yield local.__kj__.escape(i)  # noqa: F821
                    yield " is "
                    yield simple_function.evenness(i)
                    yield simple_function.half_evenness(i)
                    yield "\n"

        self.tpl = kajiki.Template(Tpl)

    def test_import(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert (
            rsp == "0 is even half of 0 is even\n"
            "1 is odd half of 1 is odd\n"
            "2 is even half of 2 is odd\n"
            "3 is odd half of 3 is odd\n"
        ), rsp


class TestInclude(TestCase):
    def setUp(self):
        class HdrUndec:
            @kajiki.expose
            def __main__():
                yield "# header\n"

        hdr = kajiki.Template(HdrUndec)

        class TplUndec:
            @kajiki.expose
            def __main__():
                yield "a\n"
                yield hdr().__main__()
                yield "b\n"

        tpl = kajiki.Template(TplUndec)
        self.tpl = tpl

    def test_include(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert rsp == "a\n# header\nb\n", rsp


class TestExtends(TestCase):
    def setUp(_self):
        class ParentTplUndec:
            @kajiki.expose
            def __main__():
                yield header()  # noqa: F821
                yield body()  # noqa: F821
                yield footer()  # noqa: F821

            @kajiki.expose
            def header():
                yield "# Header name="
                yield name  # noqa: F821
                yield "\n"

            @kajiki.expose
            def body():
                yield "## Parent Body\n"
                yield "local.id() = "
                yield local.id()  # noqa: F821
                yield "\n"
                yield "self.id() = "
                yield self.id()  # noqa: F821
                yield "\n"
                yield "child.id() = "
                yield child.id()  # noqa: F821
                yield "\n"

            @kajiki.expose
            def footer():
                yield "# Footer"
                yield "\n"

            @kajiki.expose
            def id():
                yield "parent"

        parent_tpl = kajiki.Template(ParentTplUndec)

        class MidTplUndec:
            @kajiki.expose
            def __main__():
                yield local.__kj__.extend(parent_tpl).__main__()  # noqa: F821

            @kajiki.expose
            def id():
                yield "mid"

        mid_tpl = kajiki.Template(MidTplUndec)

        class ChildTplUndec:
            @kajiki.expose
            def __main__():
                yield local.__kj__.extend(mid_tpl).__main__()  # noqa: F821

            @kajiki.expose
            def body():
                yield "## Child Body\n"
                yield parent.body()  # noqa: F821

            @kajiki.expose
            def id():
                yield "child"

        child_tpl = kajiki.Template(ChildTplUndec)
        _self.parent_tpl = parent_tpl
        _self.child_tpl = child_tpl

    def test_extends(self):
        rsp = self.child_tpl({"name": "Rick"}).render()
        assert (
            rsp == "# Header name=Rick\n"
            "## Child Body\n"
            "## Parent Body\n"
            "local.id() = parent\n"
            "self.id() = child\n"
            "child.id() = mid\n"
            "# Footer\n"
        ), rsp


class TestDynamicExtends(TestCase):
    def setUp(_self):
        class Parent0Undec:
            @kajiki.expose
            def __main__():
                yield "Parent 0"

        parent_0 = kajiki.Template(Parent0Undec)

        class Parent1Undec:
            @kajiki.expose
            def __main__():
                yield "Parent 1"

        parent_1 = kajiki.Template(Parent1Undec)

        class ChildTpl:
            @kajiki.expose
            def __main__():
                if p == 0:  # noqa: F821
                    yield local.__kj__.extend(parent_0).__main__()  # noqa: F821
                else:
                    yield local.__kj__.extend(parent_1).__main__()  # noqa: F821

        _self.child_tpl = kajiki.Template(ChildTpl)

    def test_extends(self):
        rsp = self.child_tpl({"p": 0}).render()
        assert rsp == "Parent 0", rsp
        rsp = self.child_tpl({"p": 1}).render()
        assert rsp == "Parent 1", rsp



from __future__ import annotations

import os
import sys
from pathlib import Path

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources

from kajiki.util import default_alias_for


class Loader:
    def __init__(self, reload=False):  # noqa: FBT002
        self._reload = reload
        self.modules = {}

    def import_(self, name, **kwargs):
        """Returns the template if it is already in the cache,
        else loads the template, caches it and returns it.
        """
        mod = self.modules.get(name)
        if not self._reload and mod:
            return mod
        mod = self._load(name, **kwargs)
        mod.loader = self
        self.modules[name] = mod
        return mod

    def default_alias_for(self, name):
        return default_alias_for(name)

    @property
    def load(self):
        return self.import_


class MockLoader(Loader):
    def __init__(self, modules):
        super().__init__()
        self.modules.update(modules)
        for v in self.modules.values():
            v.loader = self


class FileLoader(Loader):
    def __init__(
        self,
        path,
        reload=False,  # noqa: FBT002
        force_mode=None,
        autoescape_text=False,  # noqa: FBT002
        xml_autoblocks=None,
        **template_options,
    ):
        super().__init__(reload=reload)
        from kajiki import TextTemplate, XMLTemplate

        if isinstance(path, str):
            self.path = path.split(";")
        elif isinstance(path, Path):
            self.path = [path]
        else:
            self.path = path

        self._force_mode = force_mode
        self._autoescape_text = autoescape_text
        self._xml_autoblocks = xml_autoblocks
        self._template_options = template_options
        self.extension_map = {
            "txt": lambda **kw: TextTemplate(autoescape=self._autoescape_text, **kw),
            "xml": XMLTemplate,
            "html": lambda **kw: XMLTemplate(mode="html", **kw),
            "html5": lambda **kw: XMLTemplate(mode="html5", **kw),
        }

    def _filename(self, name: str) -> str | Path | None:
        """Get the filename of the requested resource."""
        for base in self.path:
            path = Path(base) / name
            if path.is_file():
                return path

        msg = f"{name} not found in any of {self.path}"
        raise FileNotFoundError(msg)

    def _find_resource(self, name: str) -> Path:
        """Locate the loadable resource and return a Path to it."""
        filename = self._filename(name)
        if not filename:
            msg = f"{self!r}._filename returned {filename!r}"
            raise FileNotFoundError(msg)
        path = Path(filename)
        if not path.is_file():
            msg = f"{filename} doesn't exist or isn't a file."
            raise FileNotFoundError(msg)
        return path

    def _load(self, name, encoding="utf-8", **kwargs):
        """Load a template from file."""
        from kajiki import TextTemplate, XMLTemplate

        options = self._template_options.copy()
        options.update(kwargs)

        resource = self._find_resource(name)
        source = resource.read_text(encoding=encoding)

        if self._force_mode == "text":
            return TextTemplate(
                source=source,
                filename=str(resource),
                autoescape=self._autoescape_text,
                **options,
            )

        if self._force_mode:
            return XMLTemplate(
                source=source,
                filename=str(resource),
                mode=self._force_mode,
                autoblocks=self._xml_autoblocks,
                **options,
            )

        ext = Path(resource.name).suffix.lstrip(".")
        return self.extension_map[ext](source=source, filename=str(resource), **options)


class PackageLoader(FileLoader):
    def __init__(self, reload=False, force_mode=None):  # noqa: FBT002
        super().__init__(None, reload=reload, force_mode=force_mode)

    def _find_resource(self, name):
        package, module = name.rsplit(".", 1)
        package_resource = importlib_resources.files(package)

        if package_resource.is_file():
            msg = f"{package} refers to a module, not a package."
            raise OSError(msg)

        for resource in package_resource.iterdir():
            if not resource.is_file():
                continue

            root, ext = os.path.splitext(resource.name)
            if root != module:
                continue

            for match_ext in (".xml", ".html", ".html5", ".txt"):
                if match_ext == ext:
                    return resource

        msg = f"Unknown template {name!r}"
        raise FileNotFoundError(msg)



import pytest

from kajiki.doctype import DocumentTypeDeclaration, extract_dtd

XHTML1 = (
    '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
    '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
)


@pytest.mark.parametrize(
    ("uri", "name", "rendering_mode", "stringified"),
    [
        ("", "html5", "html5", "<!DOCTYPE html>"),
        (None, "xhtml5", "xml", "<!DOCTYPE html>"),
        (
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd",
            "xhtml1transitional",
            "xml",
            XHTML1,
        ),
    ],
)
def test_dtd_by_uri(uri, name, rendering_mode, stringified):
    dtd = DocumentTypeDeclaration.by_uri[uri]
    assert dtd.name == name
    assert dtd.rendering_mode == rendering_mode
    assert str(dtd) == stringified


def test_extract_dtd():
    html = "<div>Test template</div>"
    markup = XHTML1 + html
    extracted, pos, rest = extract_dtd(markup)  # The function being tested
    assert extracted == XHTML1
    assert pos == 0
    assert rest == html
    dtd = DocumentTypeDeclaration.matching(extracted)  # Another function
    assert dtd is DocumentTypeDeclaration.by_uri["http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"]



"""End-to-end tests of Kajiki."""

import pathlib

import pytest

from kajiki.__main__ import main

DATA = pathlib.Path(__file__).parent / "data"
GOLDEN = DATA / "golden"


@pytest.mark.parametrize(
    ("args", "golden_file"),
    [
        (["-p", "kajiki_test_data.kitchensink"], "kitchensink1.html"),
        ([str(DATA / "kitchensink.html")], "kitchensink1.html"),
    ],
)
def test_golden_file(args, golden_file, capsys):
    with open(str(GOLDEN / golden_file)) as f:
        golden_data = f.read()

    main(args)

    captured = capsys.readouterr()
    assert captured.out == golden_data
    assert captured.err == ""


def test_file_not_found():
    with pytest.raises(FileNotFoundError):
        main(["/does/not/exist.txt"])


# We should be able to force a non-txt file into text mode.
def test_force_text_mode(tmpdir, capsys):
    tmpfile = str(tmpdir / "myfile.png")
    with open(tmpfile, "w") as f:
        f.write("<!DOCTYPE html>\n")
        f.write("%for i in range(10)\n")
        f.write("${i}\n")
        f.write("%end")

    main(["-m", "text", tmpfile])

    captured = capsys.readouterr()
    assert (
        captured.out
        == """<!DOCTYPE html>
0
1
2
3
4
5
6
7
8
9
"""
    )
    assert captured.err == ""



"""Comment copied from Python/compile.c:

All about a_lnotab.

c_lnotab is an array of unsigned bytes disguised as a Python string.
It is used to map bytecode offsets to source code line #s (when needed
for tracebacks).

The array is conceptually a list of
    (bytecode offset increment, line number increment)
pairs. The details are important and delicate, best illustrated by example:

    byte code offset   source code line number
       0                   1
       6                   2
      50                   7
     350                 307
     361                 308

The first trick is that these numbers aren't stored, only the increments
from one row to the next (this doesn't really work, but it's a start):

    0, 1,  6, 1,  44, 5,  300, 300,  11, 1

The second trick is that an unsigned byte can't hold negative values, or
values larger than 255, so (a) there's a deep assumption that byte code
offsets and their corresponding line #s both increase monotonically, and (b)
if at least one column jumps by more than 255 from one row to the next, more
than one pair is written to the table. In case #b, there's no way to know
from looking at the table later how many were written.	That's the delicate
part.  A user of c_lnotab desiring to find the source line number
corresponding to a bytecode address A should do something like this:

    lineno = addr = 0
    for addr_incr, line_incr in co_lnotab:
        addr += addr_incr
        if addr > A:
            return lineno
        if line_incr >= 0x80:
            line_incr -= 0x100
        lineno += line_incr

In order for this to work, when the addr field increments by more than 255,
the line # increment in each pair generated must be 0 until the remaining addr
increment is < 256.  So, in the example above, assemble_lnotab (it used
to be called com_set_lineno) should not (as was actually done until 2.2)
expand 300, 300 to 255, 255, 45, 45,
            but to 255,   0, 45, 255, 0, 45.
"""


def lnotab(pairs, first_lineno=0):
    """Yields byte integers representing the pairs of integers passed in."""
    assert first_lineno <= pairs[0][1]  # noqa: S101
    cur_byte, cur_line = 0, first_lineno
    for byte_off, line_off in pairs:
        byte_delta = byte_off - cur_byte
        line_delta = line_off - cur_line
        assert byte_delta >= 0  # noqa: S101
        while byte_delta > 255:  # noqa: PLR2004
            yield 255  # byte
            yield 0  # line
            byte_delta -= 255
        yield byte_delta
        while line_delta >= 0x80:  # noqa: PLR2004
            yield 0x7F  # line
            yield 0  # byte
            line_delta -= 0x7F
        while line_delta < -0x80:  # noqa: PLR2004
            yield 0x80  # line
            yield 0  # byte
            line_delta += 0x80
        if line_delta < 0:
            line_delta += 0x100
            assert 0x80 <= line_delta <= 0xFF  # noqa: S101, PLR2004
        yield line_delta
        cur_byte, cur_line = byte_off, line_off


def lnotab_string(pairs, first_lineno=0):
    return bytes(lnotab(pairs, first_lineno))



import sys
from unittest import TestCase

import pytest

import kajiki
from kajiki import ir


class TestBasic(TestCase):
    def setUp(self):
        self.tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.TextNode("Hello, "),
                    ir.ExprNode("name"),
                    ir.TextNode("\n"),
                )
            ]
        )

    def test(self):
        tpl = kajiki.template.from_ir(self.tpl)
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "Hello, Rick\n", rsp


class TestSwitch(TestCase):
    def setUp(self):
        self.tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.ForNode(
                        "i in range(2)",
                        ir.ExprNode("i"),
                        ir.TextNode(" is "),
                        ir.SwitchNode(
                            "i % 2",
                            ir.CaseNode("0", ir.TextNode("even\n")),
                            ir.ElseNode(ir.TextNode("odd\n")),
                        ),
                    ),
                )
            ]
        )

    def test_basic(self):
        tpl = kajiki.template.from_ir(self.tpl)
        rsp = tpl({}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestMatch:
    def setup_class(self):
        if sys.version_info < (3, 10):
            pytest.skip("pep622 unavailable before python3.10")

        self.tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.ForNode(
                        "i in range(2)",
                        ir.ExprNode("i"),
                        ir.TextNode(" is "),
                        ir.MatchNode(
                            "i % 2",
                            ir.MatchCaseNode("0", ir.TextNode("even\n")),
                            ir.MatchCaseNode("_", ir.TextNode("odd\n")),
                        ),
                    ),
                )
            ]
        )

    def test_basic(self):
        tpl = kajiki.template.from_ir(self.tpl)
        rsp = tpl({}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestFunction(TestCase):
    def setUp(self):
        self.tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "evenness(n)",
                    ir.IfNode("n % 2 == 0", ir.TextNode("even")),
                    ir.ElseNode(ir.TextNode("odd")),
                ),
                ir.DefNode(
                    "__main__()",
                    ir.ForNode(
                        "i in range(2)",
                        ir.ExprNode("i"),
                        ir.TextNode(" is "),
                        ir.ExprNode("evenness(i)"),
                        ir.TextNode("\n"),
                    ),
                ),
            ]
        )

    def test_basic(self):
        tpl = kajiki.template.from_ir(self.tpl)
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestCall(TestCase):
    def setUp(self):
        self.tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "quote(caller, speaker)",
                    ir.ForNode(
                        "i in range(2)",
                        ir.TextNode("Quoth "),
                        ir.ExprNode("speaker"),
                        ir.TextNode(', "'),
                        ir.ExprNode("caller(i)"),
                        ir.TextNode('."\n'),
                    ),
                ),
                ir.DefNode(
                    "__main__()",
                    ir.CallNode(
                        "$caller(n)",
                        "quote($caller, 'the raven')",
                        ir.TextNode("Nevermore "),
                        ir.ExprNode("n"),
                    ),
                ),
            ]
        )

    def test_basic(self):
        tpl = kajiki.template.from_ir(self.tpl)
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == 'Quoth the raven, "Nevermore 0."\n' 'Quoth the raven, "Nevermore 1."\n', rsp


class TestImport(TestCase):
    def setUp(self):
        lib = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "evenness(n)",
                    ir.IfNode("n % 2 == 0", ir.TextNode("even")),
                    ir.ElseNode(ir.TextNode("odd")),
                ),
                ir.DefNode(
                    "half_evenness(n)",
                    ir.TextNode(" half of "),
                    ir.ExprNode("n"),
                    ir.TextNode(" is "),
                    ir.ExprNode("evenness(n/2)"),
                ),
            ]
        )
        tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.ImportNode("lib.txt", "simple_function"),
                    ir.ForNode(
                        "i in range(4)",
                        ir.ExprNode("i"),
                        ir.TextNode(" is "),
                        ir.ExprNode("simple_function.evenness(i)"),
                        ir.ExprNode("simple_function.half_evenness(i)"),
                        ir.TextNode("\n"),
                    ),
                )
            ]
        )
        loader = kajiki.loader.MockLoader(
            {
                "lib.txt": kajiki.template.from_ir(lib),
                "tpl.txt": kajiki.template.from_ir(tpl),
            }
        )
        self.tpl = loader.import_("tpl.txt")

    def test_import(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert (
            rsp == "0 is even half of 0 is even\n"
            "1 is odd half of 1 is odd\n"
            "2 is even half of 2 is odd\n"
            "3 is odd half of 3 is odd\n"
        ), rsp


class TestInclude(TestCase):
    def setUp(self):
        hdr = ir.TemplateNode(defs=[ir.DefNode("__main__()", ir.TextNode("# header\n"))])
        tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.TextNode("a\n"),
                    ir.IncludeNode("hdr.txt"),
                    ir.TextNode("b\n"),
                )
            ]
        )
        loader = kajiki.loader.MockLoader(
            {
                "hdr.txt": kajiki.template.from_ir(hdr),
                "tpl.txt": kajiki.template.from_ir(tpl),
            }
        )
        self.tpl = loader.import_("tpl.txt")

    def test_include(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert rsp == "a\n# header\nb\n", rsp


class TestExtends(TestCase):
    def setUp(self):
        parent_tpl = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.ExprNode("header()"),
                    ir.ExprNode("body()"),
                    ir.ExprNode("footer()"),
                ),
                ir.DefNode(
                    "header()",
                    ir.TextNode("# Header name="),
                    ir.ExprNode("name"),
                    ir.TextNode("\n"),
                ),
                ir.DefNode(
                    "body()",
                    ir.TextNode("## Parent Body\n"),
                    ir.TextNode("local.id() = "),
                    ir.ExprNode("local.id()"),
                    ir.TextNode("\n"),
                    ir.TextNode("self.id() = "),
                    ir.ExprNode("self.id()"),
                    ir.TextNode("\n"),
                    ir.TextNode("child.id() = "),
                    ir.ExprNode("child.id()"),
                    ir.TextNode("\n"),
                ),
                ir.DefNode("footer()", ir.TextNode("# Footer\n")),
                ir.DefNode("id()", ir.TextNode("parent")),
            ]
        )
        mid_tpl = ir.TemplateNode(
            defs=[
                ir.DefNode("__main__()", ir.ExtendNode("parent.txt")),
                ir.DefNode("id()", ir.TextNode("mid")),
            ]
        )
        child_tpl = ir.TemplateNode(
            defs=[
                ir.DefNode("__main__()", ir.ExtendNode("mid.txt")),
                ir.DefNode(
                    "body()",
                    ir.TextNode("## Child Body\n"),
                    ir.ExprNode("parent.body()"),
                ),
                ir.DefNode("id()", ir.TextNode("child")),
            ]
        )
        loader = kajiki.loader.MockLoader(
            {
                "parent.txt": kajiki.template.from_ir(parent_tpl),
                "mid.txt": kajiki.template.from_ir(mid_tpl),
                "child.txt": kajiki.template.from_ir(child_tpl),
            }
        )
        self.loader = loader
        self.tpl = loader.import_("child.txt")

    def test_extends(self):
        rsp = self.tpl({"name": "Rick"}).render()
        assert (
            rsp == "# Header name=Rick\n"
            "## Child Body\n"
            "## Parent Body\n"
            "local.id() = parent\n"
            "self.id() = child\n"
            "child.id() = mid\n"
            "# Footer\n"
        ), rsp


class TestDynamicExtends(TestCase):
    def setUp(self):
        p0 = ir.TemplateNode(defs=[ir.DefNode("__main__()", ir.TextNode("Parent 0"))])
        p1 = ir.TemplateNode(defs=[ir.DefNode("__main__()", ir.TextNode("Parent 1"))])
        child = ir.TemplateNode(
            defs=[
                ir.DefNode(
                    "__main__()",
                    ir.IfNode("p==0", ir.ExtendNode("parent0.txt")),
                    ir.ElseNode(ir.ExtendNode("parent1.txt")),
                )
            ]
        )
        loader = kajiki.loader.MockLoader(
            {
                "parent0.txt": kajiki.template.from_ir(p0),
                "parent1.txt": kajiki.template.from_ir(p1),
                "child.txt": kajiki.template.from_ir(child),
            }
        )
        self.loader = loader
        self.tpl = loader.import_("child.txt")

    def test_extends(self):
        rsp = self.tpl({"p": 0}).render()
        assert rsp == "Parent 0", rsp
        rsp = self.tpl({"p": 1}).render()
        assert rsp == "Parent 1", rsp



import os
from unittest import TestCase

import pytest

from kajiki import FileLoader, MockLoader, TextTemplate


class TestBasic(TestCase):
    def test_auto_escape(self):
        tpl = TextTemplate(source="${'<h1>'}", autoescape=True)
        rsp = tpl().render()
        assert rsp == "&lt;h1&gt;", rsp

    def test_auto_escape_disable(self):
        tpl = TextTemplate(source="${literal('<h1>')}")
        rsp = tpl().render()
        assert rsp == "<h1>", rsp
        tpl = TextTemplate(source="${'<h1>'}", autoescape=False)
        rsp = tpl().render()
        assert rsp == "<h1>", rsp

    def test_expr_brace(self):
        tpl = TextTemplate(source="Hello, ${name}\n")
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "Hello, Rick\n", rsp

    def test_expr_none(self):
        tpl = TextTemplate(source="Hello, ${name}\n")
        rsp = tpl({"name": None}).render()
        assert rsp == "Hello, \n", rsp

    def test_expr_brace_complex(self):
        tpl = TextTemplate(source="Hello, ${{'name':name}['name']}\n")
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "Hello, Rick\n", rsp

    def test_expr_name(self):
        tpl = TextTemplate(source="Hello, $name\n")
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "Hello, Rick\n", rsp
        tpl = TextTemplate(source="Hello, $obj.name\n")

        class Empty:
            pass

        Empty.name = "Rick"
        rsp = tpl({"obj": Empty}).render()
        assert rsp == "Hello, Rick\n", rsp

    def test_expr_multiline(self):
        tpl = TextTemplate(
            source="""Hello, ${{'name': 'Rick',
                                               'age': 26}['name']}"""
        )
        rsp = tpl().render()
        assert rsp == "Hello, Rick", (rsp, "Hello, Rick")


class TestSwitch(TestCase):
    def test_switch(self):
        tpl = TextTemplate(
            """%for i in range(2)
$i is {%switch i % 2 %}{%case 0%}even\n{%else%}odd\n{%end%}\\
%end"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp

    def test_ljust(self):
        tpl = TextTemplate(
            """     %for i in range(2)
$i is {%switch i % 2 %}{%case 0%}even\n{%else%}odd\n{%end%}\\
%end"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp
        tpl = TextTemplate(
            """     {%-for i in range(2)%}\\
$i is {%switch i % 2 %}{%case 0%}even{%else%}odd{%end%}
    {%-end%}"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp

    def test_rstrip(self):
        tpl = TextTemplate(
            """     %for i in range(2)
$i is {%switch i % 2 %}{%case 0-%}    even\n{%else%}odd\n{%end%}\\
%end"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestFunction(TestCase):
    def test_function(self):
        tpl = TextTemplate(
            """%def evenness(n)
{%if n % 2 == 0 %}even{%else%}odd{%end%}\\
%end
%for i in range(2)
$i is ${evenness(i)}
%end
"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "0 is even\n1 is odd\n", rsp


class TestCall(TestCase):
    def test_call(self):
        tpl = TextTemplate(
            """%def quote(caller, speaker)
    %for i in range(2)
Quoth $speaker, "${caller(i)}."
    %end
%end
%call(n) quote(%caller ,'the raven')
Nevermore $n\\
%end"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == 'Quoth the raven, "Nevermore 0."\n' 'Quoth the raven, "Nevermore 1."\n', rsp


class TestImport(TestCase):
    def test_import_simple(self):
        lib = TextTemplate(
            """%def evenness(n)
%if n % 2 == 0
even\\
%else
odd\\
%end
%end
%def half_evenness(n)
 half of $n is ${evenness(n/2)}\\
%end"""
        )
        tpl = TextTemplate(
            """%import "lib.txt" as simple_function
%for i in range(4)
$i is ${simple_function.evenness(i)}${simple_function.half_evenness(i)}
%end"""
        )
        loader = MockLoader({"lib.txt": lib, "tpl.txt": tpl})
        tpl = loader.import_("tpl.txt")
        rsp = tpl({"name": "Rick"}).render()
        assert (
            rsp == "0 is even half of 0 is even\n"
            "1 is odd half of 1 is odd\n"
            "2 is even half of 2 is odd\n"
            "3 is odd half of 3 is odd\n"
        ), rsp

    def test_import_auto(self):
        lib = TextTemplate(
            """%def evenness(n)
%if n % 2 == 0
even\\
%else
odd\\
%end
%end
%def half_evenness(n)
 half of $n is ${evenness(n/2)}\\
%end"""
        )
        tpl = TextTemplate(
            """%import "lib.txt"
%for i in range(4)
$i is ${lib.evenness(i)}${lib.half_evenness(i)}
%end"""
        )
        loader = MockLoader({"lib.txt": lib, "tpl.txt": tpl})
        tpl = loader.import_("tpl.txt")
        rsp = tpl({"name": "Rick"}).render()
        assert (
            rsp == "0 is even half of 0 is even\n"
            "1 is odd half of 1 is odd\n"
            "2 is even half of 2 is odd\n"
            "3 is odd half of 3 is odd\n"
        ), rsp

    def test_include(self):
        loader = MockLoader(
            {
                "hdr.txt": TextTemplate("# header\n"),
                "tpl.txt": TextTemplate(
                    """a
%include "hdr.txt"
b
"""
                ),
            }
        )
        tpl = loader.import_("tpl.txt")
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "a\n# header\nb\n", rsp


class TestExtends(TestCase):
    def test_basic(self):
        parent = TextTemplate(
            """
%def header()
# Header name=$name
%end
%def footer()
# Footer
%end
%def body()
## Parent Body
id() = ${id()}
local.id() = ${local.id()}
self.id() = ${self.id()}
child.id() = ${child.id()}
%end
%def id()
parent\\
%end
${header()}${body()}${footer()}"""
        )
        mid = TextTemplate(
            """%extends "parent.txt"
%def id()
mid\\
%end
"""
        )
        child = TextTemplate(
            """%extends "mid.txt"
%def id()
child\\
%end
%def body()
## Child Body
${parent.body()}\\
%end
"""
        )
        loader = MockLoader({"parent.txt": parent, "mid.txt": mid, "child.txt": child})
        tpl = loader.import_("child.txt")
        rsp = tpl({"name": "Rick"}).render()
        assert (
            rsp == "# Header name=Rick\n"
            "## Child Body\n"
            "## Parent Body\n"
            "id() = child\n"
            "local.id() = parent\n"
            "self.id() = child\n"
            "child.id() = mid\n"
            "# Footer\n"
        ), rsp

    def test_dynamic(self):
        loader = MockLoader(
            {
                "parent0.txt": TextTemplate("Parent 0"),
                "parent1.txt": TextTemplate("Parent 1"),
                "child.txt": TextTemplate(
                    """%if p == 0
%extends "parent0.txt"
%else
%extends "parent1.txt"
%end
"""
                ),
            }
        )
        tpl = loader.import_("child.txt")
        rsp = tpl({"p": 0}).render()
        assert rsp == "Parent 0", rsp
        rsp = tpl({"p": 1}).render()
        assert rsp == "Parent 1", rsp

    def test_block(self):
        loader = MockLoader(
            {
                "parent.txt": TextTemplate(
                    """%def greet(name)
Hello, $name!\\
%end
%def sign(name)
Sincerely,
$name\\
%end
${greet(to)}

%block body
It was good seeing you last Friday.  Thanks for the gift!
%end

${sign(from_)}
"""
                ),
                "child.txt": TextTemplate(
                    """%extends "parent.txt"
%def greet(name)
Dear $name:\\
%end
%block body
${parent_block()}\\

And don't forget you owe me money!
%end
"""
                ),
            }
        )
        child = loader.import_("child.txt")
        rsp = child({"to": "Mark", "from_": "Rick"}).render()
        assert (
            rsp
            == """Dear Mark:
It was good seeing you last Friday.  Thanks for the gift!

And don't forget you owe me money!

Sincerely,
Rick
"""
        ), rsp


class TestClosure(TestCase):
    def test(self):
        tpl = TextTemplate(
            """%def add(x)
%def inner(y)
${x+y}\\
%end
${inner(x*2)}\\
%end
${add(5)}
"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "15\n", rsp


class TestPython(TestCase):
    def test_basic(self):
        tpl = TextTemplate(
            """%py
import os
%end
${os.path.join('a','b','c')}"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "a/b/c"

    def test_indent(self):
        tpl = TextTemplate(
            """%py
    import os
    import re
%end
${os.path.join('a','b','c')}"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "a/b/c"

    def test_short(self):
        tpl = TextTemplate(
            """%py import os
${os.path.join('a','b','c')}"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "a/b/c"

    def test_mod(self):
        tpl = TextTemplate(
            """%py% import os
%def test()
${os.path.join('a','b','c')}\\
%end
${test()}"""
        )
        rsp = tpl({"name": "Rick"}).render()
        assert rsp == "a/b/c"


class TestDebug(TestCase):
    def test_debug(self):
        loader = FileLoader(path=os.path.join(os.path.dirname(__file__), "data"))
        tpl = loader.import_("debug.txt")

        with pytest.raises(ValueError, match="Test error") as exc_info:
            tpl().render()

        # Verify we have stack trace entries in the template
        for tb_entry in exc_info.traceback:
            if tb_entry.path.name == "debug.txt":
                break
        else:
            pytest.fail("Stacktrace is all python")



from unittest import mock

import pytest

import kajiki.loader
from kajiki.__main__ import main


class MainMocks:
    def __init__(self, monkeypatch):
        mocked_render = mock.Mock(return_value="render result")
        self.render = mocked_render

        class MockedTemplate:
            def render(self, *args, **kwargs):
                return mocked_render(*args, **kwargs)

        self.template_type = mock.Mock(return_value=MockedTemplate())

        mocked_import = mock.Mock(return_value=self.template_type)
        self.import_ = mocked_import

        class MockedLoader:
            def import_(self, *args, **kwargs):
                return mocked_import(*args, **kwargs)

        self.file_loader_type = mock.Mock(return_value=MockedLoader())
        monkeypatch.setattr(kajiki.loader, "FileLoader", self.file_loader_type)

        self.package_loader_type = mock.Mock(return_value=MockedLoader())
        monkeypatch.setattr(kajiki.loader, "PackageLoader", self.package_loader_type)


@pytest.fixture
def main_mocks(monkeypatch):
    return MainMocks(monkeypatch)


@pytest.mark.parametrize(
    ("filename", "load_path"),
    [
        ("filename.txt", "."),
        ("/path/to/filename.xml", "/path/to"),
        ("some/subdir/myfile.html", "some/subdir"),
    ],
)
def test_simple_file_load(filename, load_path, capsys, main_mocks):
    main([filename])

    main_mocks.file_loader_type.assert_called_once_with(path=[load_path], force_mode=None)
    main_mocks.import_.assert_called_once_with(filename)
    main_mocks.template_type.assert_called_once_with({})
    main_mocks.render.assert_called_once_with()

    assert capsys.readouterr().out == "render result"


def test_simple_package_load(capsys, main_mocks):
    main(["-p", "my.cool.package"])

    main_mocks.package_loader_type.assert_called_once_with(force_mode=None)
    main_mocks.import_.assert_called_once_with("my.cool.package")
    main_mocks.template_type.assert_called_once_with({})
    main_mocks.render.assert_called_once_with()

    assert capsys.readouterr().out == "render result"


@mock.patch("site.addsitedir", autospec=True)
def test_package_loader_site_dirs(addsitedir, capsys, main_mocks):
    main(
        [
            "-i",
            "/usr/share/my-python-site",
            "-i",
            "relative/site/path",
            "-i",
            "another",
            "-p",
            "my.cool.package",
        ]
    )

    addsitedir.assert_has_calls(
        [
            mock.call("/usr/share/my-python-site"),
            mock.call("relative/site/path"),
            mock.call("another"),
        ]
    )

    main_mocks.package_loader_type.assert_called_once_with(force_mode=None)
    main_mocks.import_.assert_called_once_with("my.cool.package")
    main_mocks.template_type.assert_called_once_with({})
    main_mocks.render.assert_called_once_with()

    assert capsys.readouterr().out == "render result"


def test_output_to_file(tmpdir, main_mocks):
    outfile = str(tmpdir / "output_file.txt")
    main(["infile.txt", outfile])

    main_mocks.file_loader_type.assert_called_once_with(path=["."], force_mode=None)
    main_mocks.import_.assert_called_once_with("infile.txt")
    main_mocks.template_type.assert_called_once_with({})
    main_mocks.render.assert_called_once_with()

    with open(outfile) as f:
        assert f.read() == "render result"


def test_template_variables(main_mocks):
    main(["-v", "foo=bar", "-v", "baz=bip", "infile.txt"])

    main_mocks.file_loader_type.assert_called_once_with(path=["."], force_mode=None)
    main_mocks.import_.assert_called_once_with("infile.txt")
    main_mocks.template_type.assert_called_once_with(
        {
            "foo": "bar",
            "baz": "bip",
        }
    )
    main_mocks.render.assert_called_once_with()


def test_template_variables_bad(capsys):
    with pytest.raises(SystemExit) as e:
        main(["-v", "BADBADBAD", "infile.txt"])

    assert e.value.code != 0

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err.endswith("error: argument -v/--var: Expected a KEY=VALUE pair, got BADBADBAD\n")
